#include "OitoDamas.h"
#include <stdio.h>

// dados de instncia
int CDamas::N=4;
int CDamas::K=2;


CDamas::CDamas(void)
{
}

CDamas::~CDamas(void)
{
}

// Cria um objecto que e uma copia deste
TProcuraConstrutiva* CDamas::Duplicar(void)
{
	CDamas* clone = new CDamas;
	clone->damas = damas;
	clone->tabuleiro = tabuleiro;
	clone->linhas = linhas;
	clone->colunas = colunas;
	clone->diagSDIE = diagSDIE;
	clone->diagSEID = diagSEID;
	return clone;
}

// Coloca o objecto no estado inicial da procura
void CDamas::SolucaoVazia(void) {
	damas = 0; 
	tabuleiro.Count(N * N); 
	linhas.Count(N); 
	colunas.Count(N);
	diagSDIE.Count(2 * N - 1);
	diagSEID.Count(2 * N - 1);
	tabuleiro.Reset(0);
	linhas.Reset(0);
	colunas.Reset(0);
	diagSDIE.Reset(0);
	diagSEID.Reset(0);
}


// Coloca em sucessores a lista de objectos sucessores (sao alocados neste metodo e tem de ser apagados)
// O custo nao necessita de ser preenchido, caso seja sempre unitario
void CDamas::Sucessores(TVector<TProcuraConstrutiva*>&sucessores, TVector<int>&custo)
{
	int novaLinha=damas/K, i=0;
	// tentar colocar damas em todas as colunas
	if (damas % K != 0) // h damas na linha
	// gerar sucessores com damas em posies apenas superiores  ltima dama (para evitar vrias formas de se gerar o mesmo estado)
		for (i = N - 1; i > 0; i--)
			if (Casa(novaLinha, i))
				break;
	for(;i<N;i++) {
		if (Casa(novaLinha, i) == 0 && Saturacao(novaLinha, i) < K) {
			// casa possvel
			CDamas* sucessor = (CDamas*)Duplicar();
			sucessor->ColocarDama(novaLinha, i);
			sucessores.Add(sucessor);
		}
	}
	TProcuraConstrutiva::Sucessores(sucessores,custo);
}

// nmero mximo de ataques nesta casa
int CDamas::Saturacao(int linha, int coluna)
{
	int resultado = linhas[linha];
	if (resultado < colunas[coluna])
		resultado = colunas[coluna];
	if (resultado < diagSDIE[DiagonalSDIE(linha, coluna)])
		resultado = diagSDIE[DiagonalSDIE(linha, coluna)];
	if (resultado < diagSEID[DiagonalSEID(linha, coluna)])
		resultado = diagSEID[DiagonalSEID(linha, coluna)];
	return resultado;
}

void CDamas::ColocarDama(int linha, int coluna)
{
	if (Casa(linha, coluna) == 0) {
		damas++;
		linhas[linha]++;
		colunas[coluna]++;
		diagSDIE[DiagonalSDIE(linha, coluna)]++;
		diagSEID[DiagonalSEID(linha, coluna)]++;
		Casa(linha, coluna) = 1;
	}
}

void CDamas::RemoverDama(int linha, int coluna)
{
	if (Casa(linha, coluna) != 0) {
		damas--;
		linhas[linha]--;
		colunas[coluna]--;
		diagSDIE[DiagonalSDIE(linha, coluna)]--;
		diagSEID[DiagonalSEID(linha, coluna)]--;
		Casa(linha, coluna) = 0;
	}
}


// Escrever informacao de debug sobre o objecto currente
// (utilizar variavel TProcuraConstrutiva::debug para seleccionar o detalhe pretendido)
void CDamas::Debug(void)
{
	for(int i=0;i<N;i++) {
		NovaLinha();
		for(int j=0;j<N;j++) 
			printf("%c ", (Casa(i, j)?'D':'+'));
	}
	NovaLinha();
	if(damas==K*N && custo>=0) {
		printf("Custo: %d",custo);
		NovaLinha();
	}
	// mostrar linhas/colunas/diagonais atacadas
	NovaLinha();
	printf("Linhas: ");
	for (int i = 0; i < linhas.Count(); i++)
		printf("%d ", linhas[i]);
	printf("Colunas: ");
	for (int i = 0; i < colunas.Count(); i++)
		printf("%d ", colunas[i]);
	printf("DiagSEID: ");
	for (int i = 0; i < diagSEID.Count(); i++)
		printf("%d ", diagSEID[i]);
	printf("DiagSDIE: ");
	for (int i = 0; i < diagSDIE.Count(); i++)
		printf("%d ", diagSDIE[i]);
}

// e gerada uma nova solucao aleatoriamente.
void CDamas::NovaSolucao(void)
{
	SolucaoVazia();
	while (!SolucaoCompleta())
		ColocarDama(TRand::rand() % N, TRand::rand() % N);
}

// Retorna o valor da solucao completa actual.
int CDamas::Avaliar(void)
{
	TProcuraMelhorativa::Avaliar();
	custo=0;
	// calcular o numero de casas com violao de damas atacadas
	// nmero de linhas/colunas/diagonais violadas seria uma soluo mais direta
	for(int i=0;i<N;i++)
		for(int j=0;j<N;j++)
			if(Saturacao(i,j) > K)
				custo++;
	return custo;
}

// Operador de vizinhanca (apenas em solucaes completas)
// chamar este operador para actualizacao de estatisticas
void CDamas::Vizinhanca(TVector<TProcuraMelhorativa*>&vizinhos)
{
	TVector<int> linha, coluna;
	int livre = K - 1;
	// identificar as casas mais livres
	for(int i=0;i<N;i++)
		for(int j=0;j<N;j++)
			if (Casa(i, j)==0 && Saturacao(i,j) <= livre) {
				if (Saturacao(i, j) < livre) {
					linha.Count(0);
					coluna.Count(0);
					livre = Saturacao(i, j);
				}
				linha.Add(i);
				coluna.Add(j);
			}
	// mover uma dama que esteja numa casa supra-atacada
	for(int i=0;i<N;i++)
		for(int j=0;j<N;j++)
			if(Casa(i,j) && Saturacao(i, j)>K) {
				CDamas* aux = (CDamas*)Duplicar();
				aux->RemoverDama(i, j);
				// recolocar a dama numa das casas menos atacadas
				for (int w = 0; w < linha.Count(); w++) {
					CDamas* vizinho = (CDamas*)aux->Duplicar();
					vizinho->ColocarDama(linha[w], coluna[w]);
					vizinhos.Add(vizinho);
				}
				delete aux;
			}
	TProcuraMelhorativa::Vizinhanca(vizinhos);
}

// muta o estado actual (de acordo com a parametrizacao global)
void CDamas::Mutar(void)
{
	// mover damas que esteja em casas supra-atacadas, com probabilidade de 50%
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			if (Casa(i, j) && Saturacao(i,j)>K && TRand::rand() % 2 == 0) {
				int w, k;
				w = TRand::rand() % N;
				k = TRand::rand() % N;
				if (Casa(w, k) == 0) {
					RemoverDama(i, j);
					ColocarDama(w, k);
				}
			}
	TProcuraMelhorativa::Mutar();
}

// cruza os dois elementos a e b, colocando o resultado neste
void CDamas::Cruzamento(TProcuraMelhorativa *a, TProcuraMelhorativa *b)
{
	CDamas& pai = *((CDamas*)a);
	CDamas& mae = *((CDamas*)b);
	SolucaoVazia();
	// Executar os passos por ordem, at que se tenha o nmero mnimo de damas:
	// 1. coloca as damas que no pai e me esto na mesma casa
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			if (pai.Casa(i, j) && mae.Casa(i, j))
				ColocarDama(i, j);
	// 2. do pai, as restantes damas mas apenas se no ficarem em casas sobre-atacadas
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			if (pai.Casa(i, j) && Casa(i, j)==0 && Saturacao(i,j)<K)
				ColocarDama(i, j);
	// 3. da me, as restantes damas mas apenas se no ficarem em casas sobre-atacadas
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N && damas<N*K; j++)
			if (mae.Casa(i, j) && Casa(i, j) == 0 && Saturacao(i, j) < K)
				ColocarDama(i, j);
	// 4. da me, as restantes damas, mesmo que fiquem em casas sobre-atacadas
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N && damas < N * K; j++)
			if (mae.Casa(i, j) && Casa(i, j) == 0)
				ColocarDama(i, j);

	TProcuraMelhorativa::Cruzamento(a,b);
}

// retorna a distancia deste elemento ao elemento a
// (necessario para manter os elementos diversos)
int CDamas::Distancia(TProcuraMelhorativa*a)
{
	int distancia = K*N;
	CDamas& estado = *((CDamas*)a);
	// nmero de damas na mesma posio
	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
			if(Casa(i,j) && estado.Casa(i,j))
			distancia--;
	return distancia;
}
