#pragma once
#include "TProcuraMelhorativa.h"

///////////////////////////////////////////////////////////////////////////////
//	CDamas class
///////////////////////////////////////////////////////////////////////////////
//	Author: Jose Coelho
//	Description:
//    e-flio A de 21/22
//    - prespetiva construtiva: manter um tabuleiro com o nmero de damas a atacar cada linha/coluna/diagonal
//    A - procura em profundidade primeiro, sem limite, faz mais sentido, j que a soluo est no ltimo nvel
//      - ok at  instncia 10, 4
//      - a procura em largura funciona tambm bem, caso no se repitam os sucessores (colocar damas apenas aps a ltima)
//    E-flio B (mtodos apenas para o e-flio  seguinte!)
//    - heurstica: manter o maior nmero de casa sem grandes ataques (calcular mais tarde)
//    - prespetiva melhorativa: igual, mas podendo as casas terem mais que K damas a atacar
//      - Avaliao: procurar reduzir o nmero de casas de violao
//    - vizinhana mover uma dama numa casa saturada
//    B - na escalada do monte,  preferivel no mover para o primeiro melhor,  melhor fazer a vizinhana toda
//      - ok at  instncia 10,4
//    - cruzamento com posies do pai e da me
//    - mutao poder mover as damas que esto em casas muito atacadas
//    C - no algoritmo gentico, funciona sem problema
//      - aumentar a populao para 100 e exigir distncia 1 (com 100K avaliaes)
//      - resolve tudo
///////////////////////////////////////////////////////////////////////////////
class CDamas :
	public TProcuraMelhorativa
{
private:
	// acesso ao tabuleiro
	inline int &Casa(int linha, int coluna) { return tabuleiro[linha * N + coluna]; }
	int Saturacao(int linha, int coluna); // nmero mximo de ataques nesta casa
	void ColocarDama(int linha, int coluna);
	void RemoverDama(int linha, int coluna);
	int DiagonalSEID(int linha, int coluna) { return linha + coluna; }
	int DiagonalSDIE(int linha, int coluna) { return (N - 1 - coluna) + linha; }
public:
	CDamas(void);
	~CDamas(void);

	// dados de instncia
	static int N;
	static int K;

	// dados de estado
	TVector<int> tabuleiro; // damas em cada posio do tabuleiro
	TVector<int> linhas, colunas, diagSEID, diagSDIE; // nmero de damas em cada tipo de linha
	int damas; // nmero de damas colocadas

	// metodos redefinidos de TProcuraConstrutiva

	// Cria um objecto que e uma copia deste
	TProcuraConstrutiva* Duplicar(void);
	void Copiar(TProcuraConstrutiva*objecto) {
		TProcuraMelhorativa::Copiar(objecto);
		damas=((CDamas*)objecto)->damas;
		tabuleiro= ((CDamas*)objecto)->tabuleiro;
		linhas = ((CDamas*)objecto)->linhas;
		colunas = ((CDamas*)objecto)->colunas;
		diagSEID = ((CDamas*)objecto)->diagSEID;
		diagSDIE = ((CDamas*)objecto)->diagSDIE;
	}
	// Coloca o objecto no estado inicial da procura
	void SolucaoVazia(void);
	// Coloca em sucessores a lista de objectos sucessores (sao alocados neste metodo e tem de ser apagados)
	// O custo nao necessita de ser preenchido, caso seja sempre unitario
	void Sucessores(TVector<TProcuraConstrutiva*>&sucessores, TVector<int>&custo);
	// Retorna verdade caso o estado actual seja um estado objectivo
	bool SolucaoCompleta(void) { return damas==K*N; }
	// Escrever informacao de debug sobre o objecto currente
	// (utilizar variavel TProcuraConstrutiva::debug para seleccionar o detalhe pretendido)
	void Debug(void);

	// metodos redefinidos de TProcuraMelhorativa

	// e gerada uma nova solucao aleatoriamente.
	void NovaSolucao(void);
	// Retorna o valor da solucao completa actual.
	int Avaliar(void);
	// Operador de vizinhanca (apenas em solucaes completas)
	// chamar este operador para actualizacao de estatisticas
	void Vizinhanca(TVector<TProcuraMelhorativa*>&vizinhos);
	// muta o estado actual (de acordo com a parametrizacao global)
	void Mutar(void);
	// cruza os dois elementos a e b, colocando o resultado neste
	void Cruzamento(TProcuraMelhorativa*a,TProcuraMelhorativa*b);
	// retorna a distancia deste elemento ao elemento a
	// (necessario para manter os elementos diversos)
	int Distancia(TProcuraMelhorativa*a);

};
